#include <stdio.h>
#include <ctype.h>
#include <errno.h>
extern int errno;		/* just in case */
#include <sys/stat.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "mcal.h"

#ifdef CAL_VALIDUSER_KLUDGE
#include <crypt.h>
#define CAL_VALIDUSER cal_myvaliduser
#else
#define CAL_VALIDUSER cal_validuser
#endif

/* local store location */
#ifndef MSTORE_PATH
#define MSTORE_PATH "/var/calendar"
#endif

/* Autologout timer */
#define LOGINTIMEOUT 60*3
#define TIMEOUT 60*30
#define IDLETIMEOUT 60

char *flush (void);
char *lcase (char *s);
char *ucase (char *s);
char *snarf_list (char **arg);
char *snarf (char **arg);
char inchar (void);
int print_event(int uid,const CALEVENT *event);
char *strfrompos (char *original, int pos);
#ifdef CAL_VALIDUSER_KLUDGE
static bool cal_myvaliduser (const char *username, const char *password);
#endif

/* Length of literal stack (I hope 20 is plenty) */

#define LITSTKLEN 20
#define MAXCLIENTLIT 10000

/* Size of temporary buffers */
#define TMPLEN 8192
#define MAILTMPLEN TMPLEN

/* Server states */

#define LOGIN 0
#define SELECT 1
#define OPEN 2
#define LOGOUT 3
#define T 1
#define F 0

/* User types */
#define NO_USER			0
#define ANONYMOUS_USER 		1
#define AUTHENTICATED_USER 	2

/* Define initial state */
int state = LOGIN;

/* Set user type to nothing for now */
int usertype = NO_USER;

char *username=NULL;

char *lstwrn=NULL;
char *lsterr=NULL;
char *pass = NULL;		/* password */
char cmdbuf[TMPLEN];		/* command buffer */
char *tag;			/* tag portion of command */
char *cmd;			/* command portion of command */
char *arg;			/* pointer to current argument of command */

char *response = NULL;		/* command response */
int litsp = 0;			/* literal stack pointer */
char *litstk[LITSTKLEN];	/* stack to hold literals */

/* Response texts which appear in multiple places */

char *win = "%s OK %.80s completed\015\012";
char *altwin = "%s OK %.900s\015\012";
char *lose = "%s NO %.80s failed: %.900s\015\012";
char *losetry = "%s NO [TRYCREATE] %.80s failed: %.900s\015\012";
char *misarg = "%s BAD Missing required argument to %.80s\015\012";
char *badarg = "%s BAD Argument given to %.80s when none expected\015\012";
char *badseq = "%s BAD Bogus sequence in %.80s\015\012";
char *badatt = "%s BAD Bogus attribute list in %.80s\015\012";
char *argrdy = "+ Ready for argument\015\012";
char *stub   = "%s OK %.80s STUB\015\012";

/* hacked define/variables for hostname */
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 1024
#endif
char hostname [ MAXHOSTNAMELEN ];

int write_event();
char *dupstr(char *s);
void slurp (char *s,int n);
unsigned long *eventlist=NULL;
int eventlistcapacity=0;
int eventlistsize=0;
CALEVENT *event=NULL;
CALSTREAM *stream=NULL;
unsigned long uid;
int main()
{
  unsigned long i,j,k,m,uid;
  long f;
  char *tmpstr,*tmpstr_n,*s,*t,*u,*v,tmp[MAILTMPLEN];
  int year,month,day,hour,minute,second;
  datetime_t startdate,enddate;

  /* read host name using gethostname */
  if ( gethostname ( hostname, sizeof ( hostname ) ) < 0 ) {
	printf ( "could not get hostname\n" );
	exit ( 1 );
  }
  
  printf("* OK ICAP server v%s ready\r\n", VERSION);
  fflush(stdout);

 do {				/* command processing loop */
    slurp (cmdbuf,TMPLEN);	/* slurp command */
				/* no more last error or literal */
    while (litsp) free((void *) litstk[--litsp]);
				/* find end of line */
    if (!strchr (cmdbuf,'\012')) {
      if ((t = strchr (cmdbuf,' '))) *t = '\0';
      flush ();			/* flush excess */
      printf (response,t ? cmdbuf : "*");
    }
    else if (!(tag = strtok (cmdbuf," \015\012"))) {
      fputs ("* BAD Null command\015\012",stdout);
    }
    else if (strlen (tag) > 100) fputs ("* BAD Excessive tag\015\012",stdout);
    else if (!(cmd = strtok (NULL," \015\012"))) {
      printf ("%s BAD Missing command\015\012",tag);
    }
    else {			/* parse command */
      response = win;		/* set default response */
      ucase (cmd);		/* canonicalize command case */
				/* UID command? */
      if (!strcmp (cmd,"UID") && strtok (NULL," \015\012")) {
	uid = T;		/* a UID command */
	cmd[3] = ' ';		/* restore the space delimiter */
	ucase (cmd);		/* make sure command all uppercase */
      }
      else uid = 0;		/* not a UID command */
				/* snarf argument */
      arg = strtok (NULL,"\015\012");

				/* these commands always valid */
      if (!strcmp (cmd,"NOOP")) {
	if (arg) response = badarg;
      }
      else if (!strcmp (cmd,"LOGOUT")) {
	if (arg) response = badarg;
	else {			/* time to say farewell */
	  if (state == OPEN) cal_close (stream,0);
	  state = LOGOUT;
	  stream = NULL;
	  printf ("* BYE %s ICAP server terminating connection\015\012",
		  hostname);
	}
      }
      else if (!strcmp (cmd,"CAPABILITY")) {
	if (arg) response = badarg;
	else {
	  fputs ("* CAPABILITY ICAP\015\012", stdout);
	}
      }
      else if (!strcmp (cmd,"CHECK")) {
	if (arg) response = badarg;
	else {
	  /* perform maintenance functions here */
	}
      }
      else switch (state) {	/* dispatch depending upon state */
      case LOGIN:		/* waiting to get logged in */
				/* plaintext login with password */
	if (!strcmp (cmd,"LOGIN")) {
	  if (username) free ((void **) username);
	  if (pass) free ((void **) pass);
				/* two arguments */
	  if (!((username = dupstr (snarf (&arg))) &&
		(pass = dupstr (snarf (&arg))))) response = misarg;
	  else if (arg) response = badarg;
				/* see if username and password are OK */
#ifdef ALLOW_ANONYMOUS
	  else if (!strcmp (lcase(username), "anonymous")) {
	    /* anonymous access, if available */
	    state = SELECT;
	    usertype = ANONYMOUS_USER;
	    response = "%s OK %.80s ANONYMOUS completed.\015\012"; 
	  }
#endif
	  else {
            /* authenticate here */
            if ((int) CAL_VALIDUSER (username, pass)) {
              state = SELECT;
	      usertype = AUTHENTICATED_USER;
	    } else { /* not valid user */
	      response = "%s NO %.80s failed\015\012";
	    } 
	  }
	}
	/* quick stub for AUTHENTICATE command (for compatibility) */
        else if (!strcmp (cmd,"AUTHENTICATE")) {
	  if(!(t=snarf(&arg))) response = misarg;
	  else {
	    response = stub;
	  }
        }
				/* check for commands invalid in this state */
        else if (!strcmp (cmd, "CREATE") ||
		 !strcmp (cmd, "DELETE") ||	
		 !strcmp (cmd, "RENAME")) {
	  response = lose;
	}
	  break;
      
	case OPEN:		/* valid only when mailbox open */
				/* fetch mailbox attributes */
	  if (!strcmp (cmd,"FETCH") || !strcmp (cmd,"UID FETCH")) {
	    if(!(s=snarf(&arg))) response = misarg;
	    else
	      {
		t=ucase(snarf(&arg));
		if(!strcmp(t,"ICAL"))
		  {
		    cal_fetch(stream,atoi(s),&event);		
		    print_event(uid,event);
		  }
	      }
	  }

				/* close calendar store attributes */
	  else if (!strcmp (cmd,"CLOSE")) {
	    if(!(s=snarf(&arg))) {
	      /* if there is no argument, close all stores */
              if (state == OPEN) cal_close (stream, 0);
	      if (stream != NULL) response = lose;
	    } else {
	      /* close selected store */
	      response = stub;
	    }
	  }

				/* store mailbox attributes */
	  else if (!strcmp (cmd,"STORE") || !strcmp (cmd,"UID STORE")) {
				/* must have three arguments */
	    if(!(t=snarf(&arg))) response = misarg;
	    if(!(s=snarf(&arg))) response = misarg;
	    else
	      {
		if(!strcmp(s,"+FLAGS"))
		  {
		    if (arg && (s = strtok (arg," ")))
		      if(!strcmp(s,"\\DELETED")) 
			cal_remove (stream,atoi(t));
		  }
		else if(!strcmp(s,"-FLAGS"))
		  {
		    if (arg && (s = strtok (arg," ")))
		      {
		      }
		    else response = misarg;		    
		  }
		if(!strcmp(s,"ICAL"))
		  {
		    if(!(s=snarf(&arg))) response = misarg;		    		    
		    if (!arg || (*arg != '{'))
		      response = "%s BAD Missing literal in %.80s\015\012";
		    else if (!isdigit (arg[1]))
		      response = "%s BAD Missing message to %.80s\015\012";
		    else if (!(i = strtoul (arg+1,&t,10)))
		      response = "%s No Empty message to %.80s\015\012";
		    else if ((*t != '}') || t[1]) response = badarg;
		    else 
		      {              /* append the data */
                                /* tell client ready for argument */
			fputs (argrdy,stdout);
			fflush (stdout);  /* dump output buffer */
                                /* get a literal buffer */
			t = (char *) malloc (i+1);
			for (j = 0; j < i; j++) t[j] = inchar ();
			t[i] = '\0';      /* make sure tied off */
                                /* get new command tail */
			if ((j = inchar ()) == '\015') j = inchar ();
			if (j == '\012') 
			  {  /* must be end of command */
			    if (0) 
			      {
				response = lose;
			      }
			  }
			else 
			  {            /* junk after literal */
			    while (inchar () != '\012');
			    response = badarg;
			  }
			free ((void **) &t);
		      }
		  }
		
	      }
	  }
	  
	    case SELECT:              /* valid whenever logged in */
                                /* select new mailbox */
      if (!(strcmp (cmd,"SELECT") && strcmp (cmd,"EXAMINE"))) {
                                /* single argument */
          if (!(s = snarf (&arg))) response = misarg;
          else if (arg) response = badarg;
	  else 
	    {
	      tmpstr=(char *)malloc(strlen(s)+10);
	      sprintf(tmpstr,"{/mstore}%s",s);
	    }
            if (cal_open (stream,tmpstr,0)){
              state = OPEN;     /* note state open */
                                /* note readonly/readwrite */
              response = "%s OK [READ-WRITE] %s completed\015\012";
            }
            else {              /* failed */
              state = SELECT;   /* no mailbox open now */
              response = lose;  /* open failed */
            }
	    free(tmpstr);
      }

				/* CREATE calendar */
	else if (!(strcmp (cmd,"CREATE"))) {
                                /* single argument */
          if (!(s = snarf (&arg))) response = misarg;
          else if (arg) response = badarg;
	  else 
	    {
	      tmpstr = (char *)malloc(strlen(s)+10);
	      sprintf (tmpstr, "{/mstore}%s", s);
              if ((stream = cal_open (stream, tmpstr, 0))) {
	        #ifdef DEBUG
	        printf("calendar name = '%s', '%s'\n", tmpstr, strfrompos(tmpstr, 9));
		#endif
	        f = cal_create (stream, s);
	      } else {
		f = (int) NULL;
		#ifdef DEBUG
		printf ("open failed\n");
		#endif
	      }
	    }
	    if (f) {
              /*state = OPEN;*/     /* note state open */
                                /* note readonly/readwrite */
              response = "%s OK %s Completed\015\012";
            }
            else {              /* failed */
              /*state = SELECT;*/   /* no mailbox open now */
              response = lose;  /* create failed */
            }
	    free(tmpstr);
      } /* end of CREATE command */

                                /* APPEND message to mailbox */
        else if (!(strcmp (cmd,"APPEND"))) {
          u = v = NULL;          /* init flags/date */
                                /* parse mailbox name */

	  if ((s = snarf (&arg)) && arg) {
	    if (*arg == '(') {  /* parse optional flag list */
              u = ++arg;        /* pointer to flag list contents */
              while (*arg && (*arg != ')')) arg++;
              if (*arg) *arg++ = '\0';
              if (*arg == ' ') arg++;
            }
                                /* parse optional date */
            if (*arg == '"') v = snarf (&arg);
                                /* parse message */
            if (!arg || (*arg != '{'))
              response = "%s BAD Missing literal in %.80s\015\012";
            else if (!isdigit (arg[1]))
              response = "%s BAD Missing message to %.80s\015\012";
            else if (!(i = strtoul (arg+1,&t,10)))
              response = "%s No Empty message to %.80s\015\012";
            else if ((*t != '}') || t[1]) response = badarg;	    
           else {              /* append the data */
	      
                                /* tell client ready for argument */
              fputs (argrdy,stdout);
              fflush (stdout);  /* dump output buffer */
                                /* get a literal buffer */
              t = (char *) malloc (i+1);
              alarm (TIMEOUT);  /* get data under timeout */
              for (j = 0; j < i; j++) 
		{
		  t[j] = inchar ();
		  if(t[j] == '\r')
		    {
		      j--;
		      i--;
		    }
		}
              alarm (0);        /* stop timeout */
              t[i] = '\0';      /* make sure tied off */
                                /* get new command tail */
              if ((j = inchar ()) == '\015') j = inchar ();
              if (j == '\012') {/* must be end of command */
		ical_parse(&event,t,strlen(t));
		if(!strcmp(s,"NIL"))
		  cal_append(stream,"INBOX",&uid,event);
		else
		  cal_append(stream,"INBOX",&uid,event);
	      }
	      else 
		{            /* junk after literal */
		  while (inchar () != '\012');
		  response = badarg;
		}
	      free((void *) t);
	   }
	  }
	     
        }
      

                                /* list mailboxes */
        else if (!strcmp (cmd,"LIST") || !strcmp (cmd,"RLIST") ||
	         !strcmp (cmd,"LSUB")) {
                                /* get reference and mailbox argument */
          if (!((s = snarf (&arg)) && (t = snarf_list (&arg))))
            response = misarg;
          else if (arg) response = badarg;
                                /* make sure anonymous can't do bad things */
          else {
	    /* FIXME: add handler for LSUB here */
	    cal_search_range (NULL, s, t);
	  }  
	}

                                /* delete calendar store */
        else if (!strcmp (cmd,"DELETE"))  {
                                /* get reference and mailbox argument */
          if (!((s = snarf (&arg))))
            response = misarg;
          else if (arg) response = badarg;
          else {
	    /* calculate full path name */
	    tmpstr = (char *) malloc(strlen(MSTORE_PATH) + strlen(s) + 1);
	    sprintf (tmpstr, "%s/%s", MSTORE_PATH, s);
	    /* see if this can be deleted */
            if (unlink(tmpstr)) response = lose;
	  }
	}

                                /* rename calendar store */
        else if (!strcmp (cmd,"RENAME"))  {
                                /* get source and destination argument */
          if (!((s = snarf (&arg)) && (t = snarf_list (&arg))))
            response = misarg;
          else if (arg) response = badarg;
          else {
	    /* check to see if authenticated user */
	    if (usertype != AUTHENTICATED_USER) {
	    	response = lose;
	    } else { 
	      /* calculate full path name */
	      tmpstr   = (char *) malloc(strlen(MSTORE_PATH) + strlen(s) + 1);
	      tmpstr_n = (char *) malloc(strlen(MSTORE_PATH) + strlen(t) + 1);
	      sprintf (tmpstr,   "%s/%s", MSTORE_PATH, s);
	      sprintf (tmpstr_n, "%s/%s", MSTORE_PATH, t);
	      /* see if this can be deleted */
              if (rename(tmpstr, tmpstr_n)) response = lose;
	    }  
	  }
	}
	
                                /* subscribe to calendar store */
        else if (!strcmp (cmd,"SUBSCRIBE"))  {
                                /* get calendar store name */
          if (!((s = snarf (&arg))))
            response = misarg;
          else if (arg) response = badarg;
          else {
	    response = stub;
	  }
	}

                                /* unsubscribe to calendar store */
        else if (!strcmp (cmd,"UNSUBSCRIBE"))  {
                                /* get calendar store name */
          if (!((s = snarf (&arg))))
            response = misarg;
          else if (arg) response = badarg;
          else {
	    response = stub;
	  }
	}

                                /* get attributes about calendar store */
        else if (!strcmp (cmd,"ATTRIBUTES"))  {
                                /* get calendar store name */
          if (!((s = snarf (&arg)) && (t = snarf_list (&arg))))
            response = misarg;
          else if (arg) response = badarg;
          else {
	    response = stub;
	  }
	}

                                /* uid */
        else if (!strcmp (cmd,"SEARCH") || !strcmp (cmd,"UID SEARCH")){
                                /* get reference and mailbox argument */
	  
	  int deleted=0,component=0,new=0,not=0,or=0,recent=0;
	  int seen=0,tentative=0,undeleted=0,unseen=0;
	  dt_init(&startdate);
	  dt_init(&enddate);
	      while((s = snarf (&arg)))
		{
		  ucase(s);
		  if(!strcmp(s,"ALL"))
		    {
		    }
		  else if(!strcmp(s,"COMPONENT"))
		    {
		      s=ucase(snarf(&arg));
		      if(!strcmp(s,"VALARM"))
			component=1;
		      else
			component=0;
		    }
		  else if(!strcmp(s,"DELETED"))
		    {
		      deleted=1;
		    }
		  else if(!strcmp(s,"NEW"))
		    {
		      new=1;
		    }
		  else if(!strcmp(s,"NOT"))
		    {
		      not=1;
		    }
		  else if(!strcmp(s,"OR"))
		    {
		      or=1;
		    }
		  else if(!strcmp(s,"RECENT"))
		    {
		      recent=1;
		    }
		  else if(!strcmp(s,"SEEN"))
		    {
		      seen=1;
		    }
		  else if(!strcmp(s,"TENTATIVE"))
		    {
		      tentative=1;
		    }
		  else if(!strcmp(s,"UID"))
		    {
		      s=ucase(snarf(&arg));
		    }
		  else if(!strcmp(s,"UNDELETED"))
		    {
		      undeleted=1;
		    }
		  else if(!strcmp(s,"UNSEEN"))
		    {
		      unseen=1;
		    }
		  else if(!strcmp(s,"ICAL"))
		    {
		      s=ucase(snarf(&arg));		      
		      if(!strcmp(s,"DTSTART"))
			{
			  i=0;
			  s=ucase(snarf(&arg));		      
			  t=ucase(snarf(&arg));		      
			  if(strlen(t) == 16)
			    {
			      i=1;
			      if(t[16] == 'Z')
				{
				  t[16]=0;
				  second=atoi(&t[14]);
				}
			      t[14]=0x00;
			      minute=atoi(&t[12]);
			      t[12]=0x00;
			      second=atoi(&t[10]);
			      t[8]=0x00;
			    }
			  day=atoi(&t[6]);
			  t[6]=0x00;
			  month=atoi(&t[4]);
			  t[4]=0x00;
			  year=atoi(t);
			  if(!strcmp(s,"=") )
			    {
			      if(i) dt_settime(&startdate,hour,minute,second);
			      dt_setdate(&enddate, year, month, day);
			      if(i) dt_settime(&startdate,hour,minute,second);
			      dt_setdate(&enddate, year, month, day);
			    }
			  else if(!strcmp(s,">"))
			    {
			      if(i) dt_settime(&startdate,hour,minute,second);
			      dt_setdate(&startdate, year, month, day+1);
			    }
			  else if(!strcmp(s,"<"))
			    {
			      if(i) dt_settime(&enddate,hour,minute,second);
			      dt_setdate(&enddate, year, month, day-1);
			    }
			  else if(!strcmp(s,"contains"))
			    {
			    }
			  else
			    {
			      printf("BAD\r\n");
			    }
			}
		      else if(!strcmp(s,"DTEND"))
			{
			  i=0;
			  s=ucase(snarf(&arg));		      
			  t=ucase(snarf(&arg));		      
			  if(strlen(t) == 16)
			    {
			      i=1;
			      if(t[16] == 'Z')
				{
				  t[16]=0;
				  second=atoi(&t[14]);
				}
			      t[14]=0x00;
			      minute=atoi(&t[12]);
			      t[12]=0x00;
			      second=atoi(&t[10]);
			      t[8]=0x00;
			    }
			  day=atoi(&t[6]);
			  t[6]=0x00;
			  month=atoi(&t[4]);
			  t[4]=0x00;
			  year=atoi(t);
			  if(i) dt_settime(&enddate,hour,minute,second);
			  dt_setdate(&enddate, year, month, day);
			  if(!strcmp(s,"=") )
			    {
			    }
			  else if(!strcmp(s,">"))
			    {
			    }
			  else if(!strcmp(s,"<"))
			    {
			    }
			  else if(!strcmp(s,"contains"))
			    {
			    }
			  else
			    {
			      printf("BAD\r\n");
			    }
			  
			}
		    }

		  else
		    {
		      printf("BAD\r\n");
		    }
		  
		}
	      if(component) 
		cal_search_alarm(stream,&startdate);
	      else 
		cal_search_range (stream,&startdate,&enddate);
	      
	      printf("* SEARCH");	  
	      for(i=0;i<eventlistsize;i++)
		{
		  printf(" %d", (int) eventlist[i]);	  
		}
	      printf("\r\n");
	}


        else if (!strcmp (cmd,"RANGE")) {
                                /* get range */
          if (!((s = snarf (&arg)) && (t = snarf(&arg))))
            response = misarg;
          else if (arg) response = badarg;
                                /* make sure anonymous can't do bad things */

  
          else
	    {
	  day=atoi(&s[6]);
	  s[6]=0x00;
	  month=atoi(&s[4]);
	  s[4]=0x00;
	  year=atoi(s);

	  dt_init(&startdate);
	  dt_setdate(&startdate, year, month, day);

	  day=atoi(&t[6]);
	  t[6]=0x00;
	  month=atoi(&t[4]);
	  t[4]=0x00;
	  year=atoi(t);
	  dt_init(&enddate);
	  dt_setdate(&enddate, year, month, day);
	  cal_search_range (stream,&startdate,&enddate);
	  printf("* %d EXISTS\r\n",eventlistsize+1);	  
	    }
	}

      break;
      }
    		/* output most recent warning */

      if (lstwrn) printf ("* NO %s\015\012",lstwrn);
				/* get text for alternative win message now */
      if (response == altwin) cmd = lsterr;
				/* build response */
      sprintf (tmp,response,tag,cmd,"NONE");
				/* update mailbox status before response */
      fputs (tmp,stdout);	/* output response */
      
    fflush (stdout);		/* make sure output blatted */
    }
  } while (state != LOGOUT);	/* until logged out */
  exit (0);			/* all done */

}


/* Slurp a command line
 * Accepts: buffer pointer
 *	    buffer size
 */

void slurp (char *s,int n)
{
  s[--n] = '\0';		/* last buffer character is guaranteed NUL */
				/* get a command under timeout */
  while(!fgets (s,n,stdin)) {	/* read buffer */
    
    if (errno==EINTR) errno = 0;/* ignore if some interrupt */
    else {
      char *e = errno ? strerror (errno) : "command stream end of file";
      alarm (0);                /* disable all interrupts */
                                /* try to gracefully close the stream */
      if (state == OPEN) cal_close (stream,0);
      state = LOGOUT;
      stream = NULL;
      free(e);
      _exit (1);
    }
  }
}




/* Flush until newline seen
 * Returns: NULL, always
 */

char *flush (void)
{
  while (inchar () != '\012');	/* slurp until we find newline */
  response = "%s BAD Command line too long\015\012";
  return NULL;
}

/* Parse an IMAP astring
 * Accepts: pointer to argument text pointer
 *	    pointer to returned size
 *	    pointer to returned delimiter
 * Returns: argument
 */

char *parse_astring (char **arg,unsigned long *size,char *del)
{
  unsigned long i;
  char c,*s,*t,*v;
  if (!*arg) return NULL;	/* better be an argument */
  switch (**arg) {		/* see what the argument is */
  default:			/* atom */
    for (s = t = *arg, i = 0;
	 (*t > ' ') && (*t < 0x7f) && (*t != '(') && (*t != ')') &&
	 (*t != '{') && (*t != '%') && (*t != '*') && (*t != '"') &&
	 (*t != '\\'); ++t,++i);
    if ((*size = i)) break;	/* got atom if non-empty */
  case ')': case '%': case '*': case '\\': case '\0': case ' ':
   return NULL;			/* empty atom is a bogon */
  case '"':			/* hunt for trailing quote */
    for (s = t = v = *arg + 1; (c = *t++) != '"'; *v++ = c) {
      if (c == '\\') c = *t++;	/* quote next character */
				/* else must be a CHAR */
      if (!c || (c & 0x80)) return NULL;
    }
    *v = '\0';			/* tie off string */
    *size = v - s;		/* return size */
    break;

  case '{':			/* literal string */
    s = *arg + 1;		/* get size */
    if (!isdigit (*s)) return NULL;
    if ((*size = i = strtoul (s,&t,10)) > MAXCLIENTLIT) {
      /*      mm_notify (NULL,"Absurdly long client literal",ERROR); */
      return NULL;
    }
				/* validate end of literal */
    if (!t || (*t != '}') || t[1]) return NULL;
    if (litsp >= LITSTKLEN) {	/* make sure don't overflow stack */
      /*      mm_notify (NULL,"Too many literals in command",ERROR); */
      return NULL;
    }
    fputs (argrdy,stdout);	/* tell client ready for argument */
    fflush (stdout);		/* dump output buffer */
				/* get a literal buffer */
    s = v = litstk[litsp++] = (char *) malloc (i+1);
				/* get literal under timeout */
    alarm ((state != LOGIN) ? TIMEOUT : LOGINTIMEOUT);
    while (i--) *v++ = inchar ();
    alarm (0);			/* stop timeout */
    *v++ = 0x00;			/* make sure string tied off */
    				/* get new command tail */
    slurp ((*arg = v = t),TMPLEN - (t - cmdbuf));
    if (!strchr (t,'\012')) return flush ();
				/* reset strtok mechanism, tie off if done */
    if (!strtok (t,"\015\012")) *t = '\0';
    break;
  }
  if ((*del = *t)) {		/* have a delimiter? */
    *t++ = '\0';		/* yes, stomp on it */
    *arg = t;			/* update argument pointer */
  }
  else *arg = NULL;		/* no more arguments */
  return s;
}

/* Snarf a command argument (simple jacket into parse_astring())
 * Accepts: pointer to argument text pointer
 * Returns: argument
 */

char *snarf (char **arg)
{
  unsigned long i;
  char c;
  char *s = parse_astring (arg,&i,&c);
  return ((c == ' ') || !c) ? s : NULL;
}


/* Snarf a list command argument (simple jacket into parse_astring())
 * Accepts: pointer to argument text pointer
 * Returns: argument
 */

char *snarf_list (char **arg)
{
  unsigned long i;
  char c;
  char *s,*t;
  if (!*arg) return NULL;	/* better be an argument */
  switch (**arg) {
  default:			/* atom and/or wildcard chars */
    for (s = t = *arg, i = 0;
	 (*t > ' ') && (*t != '(') && (*t != ')') && (*t != '{') &&
	 (*t != '"') && (*t != '\\'); ++t,++i);
    if ((c = *t)) {		/* have a delimiter? */
      *t++ = '\0';		/* stomp on it */
      *arg = t;			/* update argument pointer */
    }
    else *arg = NULL;
    break;
  case ')': case '\\': case '\0': case ' ':
    return NULL;			/* empty name is bogus */
  case '"':			/* quoted string? */
  case '{':			/* or literal? */
    s = parse_astring (arg,&i,&c);
    break;
  }
  return ((c == ' ') || !c) ? s : NULL;
}

char *ucase (char *s)
{
  char *t;
                                /* if lowercase covert to upper */
  for (t = s; *t; t++) if (!(*t & 0x80) && islower (*t)) *t = toupper (*t);
  return s;                     /* return string */
}


/* Convert string to all lowercase
 * Accepts: string pointer
 * Returns: string pointer
 */

char *lcase (char *s)
{
  char *t;
                                /* if uppercase covert to lower */
  for (t = s; *t; t++) if (!(*t & 0x80) && isupper (*t)) *t = tolower (*t);
  return s;                     /* return string */
}

void cc_vlog(const char *fmt, va_list ap)
{
}
void cc_vdlog(const char *fmt, va_list ap)
{
}
void cc_login(const char **user, const char **password)
{
  *user=username;
  *password=pass;
}
void cc_searched(unsigned long id)
{

  if(eventlistsize == eventlistcapacity)
    {
      if(eventlistcapacity==0)
        {
          eventlistcapacity=8;
        }
      eventlistcapacity <<=1;
    }
  eventlist=realloc(eventlist,eventlistcapacity*sizeof(*eventlist));
  eventlist[eventlistsize++]=id;

}

char *dupstr(char *s)
{
  return(s ? strdup((const char *)s) : NULL); 
}

char inchar (void)
{
  int c;
  while ((c = getchar ()) == EOF) {
    if (errno==EINTR) errno = 0;/* ignore if some interrupt */
    else {
      char *e = errno ? strerror (errno) : "command stream end of file";
                                /* try to gracefully close the stream */
      if (state == OPEN) cal_close (stream,0);
      state = LOGOUT;
      stream = NULL;
      free (e);
      _exit (1);
    }
  }
  return c;
}

int print_event(int uid,const CALEVENT *event)
{
        FILE            *tmp;
        char            *buf;

        if ((tmp = icalout_begin()) == NULL)
	  return false;
        if (!icalout_event(tmp, event))
	  return false;
        if ((buf = icalout_end(tmp)) == NULL)
	  return false;

	printf("* %d FETCH ICAL {%d}\r\n",uid,strlen(buf)-2);
	printf("%s",buf);
        free(buf);

        return 1;
}

/* this is a kludge until cal_validuser is implemented in the libmcal API */
#ifdef CAL_VALIDUSER_KLUDGE

#ifndef MPASSWD_PATH
#define MPASSWD_PATH "/etc/mpasswd"
#endif

bool cal_myvaliduser (const char *username, const char *password)
{
  FILE *mpasswd;
  char line[1000];
  char *musername,*mpassword;

  mpasswd=fopen(MPASSWD_PATH,"r");
  if (!mpasswd) {
    printf("Error! couldn't open mpasswd file!\n");
    exit(1);
  }
  while (fgets(line,900,mpasswd)) {
    if (line[strlen(line)-1]=='\n') line[strlen(line)-1]=0x00;
    musername=line;
    mpassword=strchr(line,':');
    *mpassword=0x00;
    mpassword++;
    if (!strcmp(username,musername)) {
      if (!strcmp(crypt(password,mpassword),mpassword)) {
        fclose(mpasswd);
	return true;
      } else {
        fclose(mpasswd);
	return false;
      }
    }
  }
  fclose(mpasswd);
  return false;
}

#endif


/* Get a substring, starting at a particular string position
 * Accepts: string, position
 * Returns: new string
 */

char *strfrompos (char *original, int pos)
{
  char *working = NULL;
  int i;

  /* if the position is out of bounds, return original */
  if ((pos > strlen(original)) || (pos < 0)) return original;

  /* malloc proper size */
  working = malloc ((strlen(original) - pos) + 1);

  /* loop through char positions */
  for (i=pos;i<strlen(original);i++) {
    working[i-pos] = original[i];
    printf ("i = %d, working[%d] = '%c', original[i] = '%c'\n",
      i, (i-pos), working[i-pos], original[i]);
  }
  working[(strlen(original) - pos)] = '\0';
  
  return strdup(working);
}
